/* 
 * CSE142, 99 Winter
 * Homework 5
 * Valentine's Day Card Maker
 *
 * Author:     Brad Fitzpatrick
 * Email:      bradfitz@bradfitz.com
 * Student ID: 9825938
 * Section:    AH
 *	
 */

/* NOTE TO READER:
 *
 * I started off doing this project intending to just do a few things
 * over the requirements.  I ended up doing a lot more.  As I found out,
 * GP142 is really slow.  Towards the end of the week I started
 * playing with the Win32 API to do cool little tricks and speed 
 * things up.  HOWEVER--- everything still works with WIN32_ENHANCED
 * turned off.  It's just slower and doesn't do as much.  I didn't
 * get around to optimizing half of what I want to, but oh well.
 * It was fun.
 *
 * Compile with EMAIL_SHELL_EXECUTE on if you want the email button
 * to launch a new email window.  The compiler on the turn-in page
 * doesn't know ShellExecute or something, so it's off for now.
 * It's a lot more fun with it on.  :)
 */

/*************************************************************************/
/***************************** MISCELLANEOUS *****************************/
/*************************************************************************/

#define WIN32_ENHANCED       1  /* When on, Win32 API is used when desireable */
#define EMAIL_SHELL_EXECUTE  0

#include "gp142.h"
#include <math.h>
#include <time.h>
#include <string.h>
#include <stdio.h>

#define NEW(type) (type *)malloc(sizeof(type))

/* If Win32, let us have access to the fun stuff. */
#if defined(WINDOWS) && WIN32_ENHANCED
extern HDC offscreen_DC, the_hdc;
extern COLORREF gRGBColors[MAX_COLORS];
extern HWND hDrawWnd;
static int gXCenter = GP142_XMAX - 3, gYCenter = GP142_YMAX - 2;
#endif

/*************************************************************************/
/************************** CONSTANTS ************************************/
/*************************************************************************/

#define TRUE    		1
#define FALSE   		0

#define BUTTON_HEIGHT  	30
#define BUTTON_WIDTH 	100

#define FILL 0
#define NOFILL 1

#define CARD_Y       220
#define CARD_X      -60
#define CARD_WIDTH   220
#define CARD_HEIGHT  380
#define CARD_ZRAISE  30   /* Shadow size */

/* Middle of card */
#define CARD_MX      (CARD_X + CARD_WIDTH / 2)
#define CARD_MY      (CARD_Y - CARD_HEIGHT / 2)

#define SHADOW_Z 5
#define SHADOW_COLOR MED_GRAY
#define BUTTON_COLOR LT_GRAY

#define SIDE_FRONT 0
#define SIDE_INSIDE 1
#define SIDE_BACK 2

#define EVENT_NONE          0
#define EVENT_CHANGE_BG     1
#define EVENT_FLIP_SIDE     2
#define EVENT_ADD_OBJECT    3
#define EVENT_ADD_TEXT      4
#define EVENT_CHANGE_COLOR  5
#define EVENT_SELECT        6
#define EVENT_DELETE        7
#define EVENT_CHANGE_OBJECT 8
#define EVENT_MOVE          9
#define EVENT_COLOR         10
#define EVENT_CHANGE_SIZE   11
#define EVENT_CLONE         12
#define EVENT_KILL_ALL      13
#define EVENT_EMAIL         14

#define SEL_PAD       2      /* padding area around chasing ants */
#define FLIP_SPEED    0.06
#define MOVE_SPEED    0.15
#define PI			  3.14159
#define COS_HALF_PI   0.70710

#define STATUS_NONE          0
#define STATUS_ADD_OBJECT    1
#define STATUS_ADD_TEXT      2
#define STATUS_SELECT        3
#define STATUS_MOVE          4

#define MAX_SHAPES      4
#define SHAPE_HEART     0
#define SHAPE_STAR      1
#define SHAPE_BALLOON   2
#define SHAPE_DUB       3

#define OBJ_TEXT       1
#define OBJ_SHAPE      2
#define OBJ_EXPLOSION  3   /* for when an item is deleted */

#define SIZE_SMALL       1
#define SIZE_SMALL_MED   2
#define SIZE_MEDIUM      3
#define SIZE_MED_LARGE   4
#define SIZE_LARGE       5

#define SELECT_RADIUS    30   /* how close you have to click to object base to select it */

#define MAX_STRING_SIZE  500  /* if user enters message > this value, buffer overrun!  crash! */

#define DEFAULT_MESSAGE  "Happy Valentine's Day"

/* Variables for the status bar at the bottom */
#define X_SIDE_PAD    5
#define Y_SIDE_PAD    5
#define STATUS_HEIGHT 20

/*************************************************************************/
/******************************** STRUCTURES *****************************/
/*************************************************************************/

/* A button */
typedef struct button_s
{
	struct button_s *next;  /* Bi-directional linked list */
	struct button_s *prev;

	char accel_key;  /* accelerator key (keyboard shorcut) */

	int color;
	int x1, y1, x2, y2;
	char* label;
	int event;
	int subevent;
	int pressed;
	int enabled;      /* flag; enabled or not?  updated by draw_button, read by handle_button */
	void *enabledvar; /* pointer to spot that must be non-zero for this button to be enabled */
	int lightstatus;  /* for what status does this button stay lit? */
} button_t;

/* Button collection */
typedef struct buttons_s
{
	int count;
	struct button_s *first;
	struct button_s *last;
} buttons_t;

/* An object (shape or text) */
typedef struct obj_s
{
	struct obj_s *next;  /* Bi-directional linked list */
	struct obj_s *prev;

	int type;
	int shape;
	int cycle;
	int color;
	int size;

	int recalc_coords;      /* Flag: when on, drawing function 
	                         * should calculate cx1,cxy1,etc boundaries
							 * for use later   */
	int cx1, cx2, cy1, cy2; /* Edge boundaries for selection drawing & selection */

	char *text;
	int x, y, side;

	int from_x, from_y, to_x, to_y;  /* Where it's sliding from and to */
	double acnt;                     /* Percent from 0.0 to 1.0 on sliding line */

} obj_t;

/* Object collection */
typedef struct objs_s
{
	int count;
	struct obj_s *first;
	struct obj_s *last;
} objs_t;

/* Structure of a side of a card (front, inside, back) */
typedef struct side_s
{
	int x1, y1, x2, y2;
	int color;
} side_t;

/* Structure containing all "global" variables for an instance
   of the card creator.  Most major functions take a pointer
   to the instance structure
*/
typedef struct
{
	int status;        /* status mode (see the #define's) */
	char *statustext;  /* Text in status bar */

	int sel_cycle;      /* A counter to say where the chasing ants are */
	int last_gp_event;  
	int side;       /* Current side we're on */
	int side_to;    /* Side we're flipping to */
	int side_show;  /* Side that the textual display says */
	char *sidetext;

	double sacnt;   /* Side Animation Count (0 to 1) */

	struct obj_s *sel_object;  /* currently selected object */

	struct side_s *sides;      /* Actually a pointer to a 3 element array */
	struct buttons_s *buttons; /* linked list */
	struct objs_s *objects; /* linked list */
} instvars_t;


/*************************************************************************/
/***********************  FUNCTION PROTOTYPES ****************************/
/*************************************************************************/

int get_fontsize(int size);
void delete_object (instvars_t *ins, obj_t *selobject);
void add_button (buttons_t *buttons, int x1, int y1, int x2, int y2, int event, int subevent, int lightstatus, void *enabledvar, char* label, char accel_key);
void add_sidebar_button (buttons_t *buttons, int pos, int event, int subevent, int lightstatus, void *enabledvar, char *label, char accel_key);
void set_object_coords (obj_t *obj, int cx1, int cy1, int cx2, int cy2);
void text_size (int *width, int *height, int fontsize, const char *string);
void assign_string (char **string_ptr, const char *source);
void handle_key (instvars_t *ins, int key);
void set_status (instvars_t *ins, char *statustext);
void free_all (instvars_t *ins);
void draw_blank_screen (instvars_t *ins);
void draw_buttons(instvars_t *ins);
void draw_card (instvars_t *ins);
void draw_sel_border (instvars_t *ins);
void draw_heart (instvars_t *ins, obj_t* obj);
void draw_star (instvars_t *ins, obj_t *obj);
void draw_balloon (instvars_t *ins, obj_t *obj);
void draw_dub (instvars_t *ins, obj_t *obj);
void handle_event (instvars_t *ins, int event, int subevent, int mouse_x, int mouse_y);
void handle_mouse_click(int mouse_x, int mouse_y, instvars_t *ins);
void handle_button(int x, int y, int *event, int *subevent, instvars_t *ins);
obj_t* clone_object(obj_t* source);
obj_t* create_object(int x, int y, instvars_t *ins);
void add_object (instvars_t *ins, obj_t *obj);
void draw_object (instvars_t *ins, obj_t *obj);
void draw_objects (instvars_t *ins);


/*************************************************************************/
/****************************** MAIN *************************************/
/*************************************************************************/

int main(void)
{
	int i, j;
	int quit = FALSE;              
	int mouse_x, mouse_y;          
	char key_hit;
	instvars_t *ins;

	side_t sides[3];
	time_t t;

	/* Reordering of the 24 standard colors to a more "purty" order */
	int color_array[24] = {
		0,	22,	23,	1, 
		2,	6,	8,	10, 
		9,	4,	7,	11, 
		19,	3,	20,	12,
		13, 18,	5,	16,
		14,	17,	21, 15
	};

#if defined(WINDOWS) && WIN32_ENHANCED
	/* Will need these for centering the window on the screen */
	RECT rctScreen, rctWindow;
#endif

	/*
	 * INITIALIZATIONS
	 */

	/* set random seed */
	srand (time(&t));	

	/* GP142 initializations */
	GP142_open();
	GP142_show_text(0);
	GP142_logging(LOG_OFF);
	GP142_clear();
	GP142_animate(ANI_RUN);

#if defined(WINDOWS) && WIN32_ENHANCED
	/* Since we can, change the application's title bar to something cool. */
	SetWindowText(hDrawWnd, "Brad's Card Maker");

	/* Oh, let's center the window on the visible workspace while 
	 * we're at it.  GP142 makes the windows at (0,0) but I have my
	 * taskbar on the left side of the screen so it's always underneith.
	 * Very annoying.  Let's fix it so it's nice wherever people's
	 * taskbar is.  Unfortunately it's gonna flicker because GP142_open
	 * shows the window right after it makes it.  You'll see the
	 * window in the upper left and the text window and then the text
	 * window go away and the draw window center itself.
	 */
	SystemParametersInfo(
		SPI_GETWORKAREA,
		0,
		&rctScreen,
		0
	);
	GetWindowRect(
		hDrawWnd,
		&rctWindow
	);
	MoveWindow(
		hDrawWnd,
		(rctScreen.right+rctScreen.left-(rctWindow.right-rctWindow.left))/2,
		(rctScreen.bottom+rctScreen.left-(rctWindow.bottom-rctWindow.left))/3,
		(rctWindow.right-rctWindow.left),
		(rctWindow.bottom-rctWindow.left),
		1
	);

	/* Note: we could've just wrote our own GP142_open, but it's probably 
	 * best to screw with as little as possible from GP142.
	 */

#endif

	/* Make the instance structure to hold passed-along variables */
	ins = NEW(instvars_t);
	ins->side = SIDE_FRONT;
	ins->side_to = SIDE_FRONT;
	ins->side_show = SIDE_FRONT;
	ins->sacnt = 0.0;
	ins->status = STATUS_NONE;
	ins->sel_object = NULL;
	ins->statustext = NULL;
	ins->sel_cycle = 0;
	set_status(ins, "Welcome to Brad's Card Maker!");
	ins->sidetext = NULL;
	assign_string(&ins->sidetext, "Front of Card");

	/* Make the sides of the cards */
	for (i=SIDE_FRONT; i<=SIDE_BACK; i++)
	{
		sides[i].x1 = CARD_X;
		sides[i].x2 = CARD_X+CARD_WIDTH;
		sides[i].y1 = CARD_Y;
		sides[i].y2 = CARD_Y-CARD_HEIGHT;
		sides[i].color = WHITE;
	}
	sides[SIDE_INSIDE].x1 = CARD_X-CARD_WIDTH/2;
	sides[SIDE_INSIDE].x2 = CARD_X+3*CARD_WIDTH/2;
	ins->sides = sides;

	/* Create the objects */
	ins->objects = NEW(objs_t);
	ins->objects->count = 0;
	ins->objects->first = NULL;
	ins->objects->last = NULL;

	/* Create the buttons */
	ins->buttons = NEW(buttons_t);
	ins->buttons->count = 0;
	ins->buttons->first = NULL;
	ins->buttons->last = NULL;

	/* Don't really need these anymore:  (handled by other modes/events now)
		add_sidebar_button(ins->buttons, 0, EVENT_CHANGE_BG, 0, 0, NULL, "Card Color");
		add_sidebar_button(ins->buttons, 3, EVENT_SELECT, 0, STATUS_SELECT, &ins->objects->count, "Select Object");
		add_sidebar_button(ins->buttons, 6, EVENT_CHANGE_COLOR, 0, 0, &ins->sel_object, "Change Color");  
	*/

	add_sidebar_button(ins->buttons, 0, EVENT_ADD_OBJECT, 0, STATUS_ADD_OBJECT, NULL, "(A)dd Shape", 'A');
	add_sidebar_button(ins->buttons, 1, EVENT_ADD_TEXT, 0, STATUS_ADD_TEXT, NULL, "Add (T)ext", 'T');
	add_sidebar_button(ins->buttons, 2, EVENT_MOVE, 0, STATUS_MOVE, &ins->sel_object, "(M)ove Object", 'M');
	add_sidebar_button(ins->buttons, 3, EVENT_CHANGE_OBJECT, 0, 0, &ins->sel_object, "C(h)ange Object", 'H');
	add_sidebar_button(ins->buttons, 4, EVENT_DELETE, 0, 0, &ins->sel_object, "(D)elete Object", 'D');
	add_sidebar_button(ins->buttons, 5, EVENT_KILL_ALL, 0, 0, NULL, "(K)ill All", 'K');
	add_sidebar_button(ins->buttons, 6, EVENT_CLONE, 0, 0, &ins->sel_object, "(C)lone Object", 'C');

	/* Here I use standard sidebar button add functions, but then change their 
	 * values after they're constructed.
	 */
	add_sidebar_button(ins->buttons, 7, EVENT_CHANGE_SIZE, -1, 0, &ins->sel_object, "Size -", '-');
	ins->buttons->last->x2 = (ins->buttons->last->x2+ins->buttons->last->x1)/2 - 5;

	add_sidebar_button(ins->buttons, 7, EVENT_CHANGE_SIZE, 1, 0, &ins->sel_object, "Size +", '+');
	ins->buttons->last->x1 = ins->buttons->last->prev->x2 + 5;

	/* Make the page flipping buttons manually... */
	add_button(ins->buttons, CARD_X+CARD_WIDTH/2+5, CARD_Y-CARD_HEIGHT-CARD_ZRAISE/2, CARD_X+CARD_WIDTH/2+40, CARD_Y-CARD_HEIGHT-CARD_ZRAISE/2-30, EVENT_FLIP_SIDE, 1, 0, NULL, "-->", '.');
	add_button(ins->buttons, CARD_X+CARD_WIDTH/2+5-40, CARD_Y-CARD_HEIGHT-CARD_ZRAISE/2, CARD_X+CARD_WIDTH/2, CARD_Y-CARD_HEIGHT-CARD_ZRAISE/2-30, EVENT_FLIP_SIDE, -1, 0, NULL,"<--", ',');

	/* Make the email Brad button */
	add_button(ins->buttons, GP142_XMAX-X_SIDE_PAD-150, -GP142_YMAX+STATUS_HEIGHT+45, GP142_XMAX-X_SIDE_PAD, -GP142_YMAX+STATUS_HEIGHT+15, EVENT_EMAIL, 0, 0, NULL, "bradfitz@bradfitz.com", '@');


	/* Make the color pallette buttons */
	j = -1;
	for (i=0; i<MAX_COLORS; i++)
	{
		/* If every forth row, move down a row */
		if (! (i % 4)) j++;

		add_button(
			ins->buttons, 
			-GP142_XMAX+10+25*(i%4),
			-GP142_YMAX+190-j*25,
			-GP142_XMAX+10+25*(i%4)+23,
			-GP142_YMAX+190-23-j*25, 
			EVENT_COLOR, 
			color_array[i],   /* sub-event is color number */
			0,                /* no associated status mode */
			NULL,             /* no enabled condition */
			"",               /* no label */
			'\0'              /* no accelerator key */
		);

		/* Now, take advantage of the fact that we know how to get 
		 * to the button we just created to change it's color, which 
		 * isn't part of the constructor.
		 */
		ins->buttons->last->color = color_array[i];
	}

	/* Event Loop */
	while (!quit)
	{
		draw_blank_screen(ins);
		draw_buttons(ins);
		draw_card(ins);
		draw_sel_border(ins);			

		ins->last_gp_event = GP142_await_event(&mouse_x, &mouse_y, &key_hit);
		
		switch (ins->last_gp_event)
		{
			case GP142_QUIT:
				quit = TRUE;
				break;
				
			case GP142_KBD:
				handle_key(ins, key_hit);
				break;
				
			case GP142_MOUSE:
				handle_mouse_click(mouse_x, mouse_y, ins);			
				break;
				
			case GP142_PERIODIC:
				break;

			default:
				break;
		} 
	}
	
	free_all(ins);  /* Free up all memory we've allocated */
	GP142_close();  /* Close windows */
	return 0;
}


/*************************************************************************/
/***************************** FUNCTIONS *********************************/
/*************************************************************************/


int get_fontsize(int size)
{
	/* These are totally arbitrary.  Change at will.... */
	switch (size)
	{
		case SIZE_SMALL:     return 14;
		case SIZE_SMALL_MED: return 18;
		case SIZE_MEDIUM:    return 22;
		case SIZE_MED_LARGE: return 30;
		case SIZE_LARGE:     return 40;
	}
	return 22;
}

void delete_object (instvars_t *ins, obj_t *selobject)
{
	if (ins->objects->last == selobject)
		ins->objects->last = selobject->prev;

	if (ins->objects->first == selobject)
		ins->objects->first = selobject->next;

	if (selobject->prev)
		selobject->prev->next = selobject->next;

	if (selobject->next)
		selobject->next->prev = selobject->prev;

	ins->objects->count--;

	if (selobject->text)
		free(selobject->text);

	free(selobject);
}

void add_button (buttons_t *buttons, int x1, int y1, int x2, int y2, int event, int subevent, int lightstatus, void *enabledvar, char* label, char accel_key)
{
	/* Create new button */
	button_t *button = NEW(button_t);
	button->x1 = x1;
	button->x2 = x2;
	button->y1 = y1;
	button->y2 = y2;
	button->label = (char *)malloc(sizeof(char)*(strlen(label)+1));
	strcpy(button->label, label);
	button->event = event;
	button->subevent = subevent;
	button->lightstatus = lightstatus;
	button->pressed = 0;
	button->enabled = 1;
	button->enabledvar = enabledvar;
	button->color = BUTTON_COLOR;
	button->accel_key = accel_key;

	/* Add button to list */
	buttons->count++;
	button->next = NULL;

	if (buttons->count == 1)
	{
		buttons->first = button;
		buttons->last = button;
		button->prev = NULL;
	}
	else
	{
		buttons->last->next = button;
		button->prev = buttons->last;
		buttons->last = button;
	}
}

void add_sidebar_button (buttons_t *buttons, int pos, int event, int subevent, int lightstatus, void *enabledvar, char *label, char accel_key)
{
	add_button(
		buttons, 
		-GP142_XMAX + 10,
		GP142_YMAX - pos * (BUTTON_HEIGHT + 5) - 10,
		-GP142_XMAX + 10 + BUTTON_WIDTH,
		GP142_YMAX - pos * (BUTTON_HEIGHT + 5) - BUTTON_HEIGHT - 10,
		event,
		subevent,
		lightstatus,
		enabledvar,
		label,
		accel_key
	);
}

/* Set an object's boundaries */
void set_object_coords (obj_t *obj, int cx1, int cy1, int cx2, int cy2)
{
	obj->cx1 = cx1;
	obj->cx2 = cx2;
	obj->cy1 = cy1;
	obj->cy2 = cy2;
	obj->recalc_coords = 0;   /* set flag off, so less CPU wasted */
}

/* Return the width and height of a text string at a certain font size */
void text_size (int *width, int *height, int fontsize, const char *string)
{
#if defined(WINDOWS) && WIN32_ENHANCED
	/* If Windows, use the Win32 API to see text size */
    HFONT hFont, hOldFont;
    SIZE textSize;

	/* Ensure Valid Font Size */
    if (fontsize < 0) fontsize = -fontsize;
    if (fontsize > 127) fontsize = 127;

    hFont = CreateFont(-fontsize,  /* negative means point size */
        0,          /* let Windows choose width */
        0,          /* escapement, in 10ths of degrees */
        0,          /* orientation, in 10ths of degrees */
        FW_NORMAL,  /* normal weight */
        FALSE,      /* no italic */
        FALSE,      /* no underline */
        FALSE,      /* no strikeout */
        ANSI_CHARSET,   /* use ANSI character set */
        OUT_DEFAULT_PRECIS, /* output precision */
        CLIP_DEFAULT_PRECIS,    /* clip precision */
        DEFAULT_QUALITY,    /* quality */
        DEFAULT_PITCH | FF_DONTCARE,    /* pitch and family */
        "GP142-font");

    hOldFont = SelectObject(offscreen_DC, hFont);

    GetTextExtentPoint(offscreen_DC,
        string,
        strlen(string),
        &textSize);

	*width = textSize.cx;
	*height = textSize.cy;

#else
	/* If not Windows, do some rough calculations */
	*width = (int) (0.5 * fontsize * strlen(string));
	*height = fontsize;
#endif

return;

}


/* Given an address of a char pointer, allocates memory for a new copy
 * of source, copies source in new area, and points string_ptr at the
 * new address
 */
void assign_string (char **string_ptr, const char *source)
{
	/* TODO: verify that this logic is correct */
	char *newstr;

	if (*string_ptr)
		free(*string_ptr);  

	if (source)
	{
		newstr = (char *)malloc(sizeof(char)*(1+strlen(source)));
		strcpy(newstr, source);
		*string_ptr = newstr;
	}
	else
	{
		*string_ptr = NULL;
	}
}

void handle_key (instvars_t *ins, int key)
{
	button_t *b = ins->buttons->first;

	while (b)
	{
		if (b->enabled && b->accel_key==toupper(key))
		{
			b->pressed = 1;
			handle_event(ins, b->event, b->subevent, 0, 0);
			return;
		}
		b = b->next;
	}
}

void set_status (instvars_t *ins, char *statustext)
{
	assign_string (&ins->statustext, statustext);
}

void free_all (instvars_t *ins)
{
	obj_t *o, *of;     /* object and object to free */
	button_t *b, *bf;  /* button and button to free */

	/* Free all the objects/shapes */
	o = ins->objects->first;
	while (o)
	{
		of = o;
		o = o->next;

		if (of->text) free(of->text);
		free(of);
	}

	/* Free all the buttons */
	b = ins->buttons->first;
	while (b)
	{
		bf = b;
		b = b->next;

		if (bf->label) free(bf->label);
		free(bf);
	}

	/* Free the structure holding the instance variables */
	free(ins);
}

void draw_blank_screen (instvars_t *ins) 
{
	GP142_rectangleXY (LT_GRAY, -GP142_XMAX, -GP142_YMAX, GP142_XMAX, GP142_YMAX, FILL);

	/* Make a status bar tray */
	/* Dark, upper bevel */
	GP142_lineXY (MED_GRAY, -GP142_XMAX+X_SIDE_PAD, -GP142_YMAX+Y_SIDE_PAD, -GP142_XMAX+X_SIDE_PAD, -GP142_YMAX+Y_SIDE_PAD+STATUS_HEIGHT, 1);
	GP142_lineXY (MED_GRAY, -GP142_XMAX+X_SIDE_PAD, -GP142_YMAX+Y_SIDE_PAD+STATUS_HEIGHT, GP142_XMAX-X_SIDE_PAD, -GP142_YMAX+Y_SIDE_PAD+STATUS_HEIGHT, 1);
	/* Bright, lower bevel */
	GP142_lineXY (WHITE, GP142_XMAX-X_SIDE_PAD, -GP142_YMAX+Y_SIDE_PAD+STATUS_HEIGHT-1, GP142_XMAX-X_SIDE_PAD, -GP142_YMAX+Y_SIDE_PAD, 1);
	GP142_lineXY (WHITE, GP142_XMAX-X_SIDE_PAD, -GP142_YMAX+Y_SIDE_PAD, -GP142_XMAX+X_SIDE_PAD-1, -GP142_YMAX+Y_SIDE_PAD, 1);

	GP142_printfXY (BLACK, -GP142_XMAX+X_SIDE_PAD+3, -GP142_YMAX+Y_SIDE_PAD+3, 14, ins->statustext);
}


/* 
 * Draws the buttons 
 */
void draw_buttons(instvars_t *ins)
{
	button_t *b = ins->buttons->first;
	int i = 0;
	int top_col, bot_col;
	int shift;
	int pressed;
	int innercolor;

#if defined(WINDOWS) && WIN32_ENHANCED

	HBRUSH hBrush, hOldBrush = NULL;
    HPEN hPen, hBlackPen, hWhitePen, hOldPen = NULL;

	hWhitePen = CreatePen(PS_SOLID, 1, gRGBColors[WHITE]);
	hBlackPen = CreatePen(PS_SOLID, 1, gRGBColors[BLACK]);

#endif

	while (b)
	{
		pressed = b->pressed || (b->lightstatus && b->lightstatus == ins->status);
		/* Flip bevel colors when pressed down */
		top_col = pressed ? BLACK : WHITE;
		bot_col = pressed ? WHITE : BLACK;
		shift = pressed ? 1 : 0;
		
		/* Button is enabled if there's no enabled condition or if the value
		   that the enabled condition pointer points to is non-zero  */
		b->enabled = (b->enabledvar == NULL) || *(int *)(b->enabledvar);

		innercolor = (b->lightstatus && b->lightstatus == ins->status) ? MED_GRAY : b->color;

#if defined(WINDOWS) && WIN32_ENHANCED

		hPen = CreatePen(PS_SOLID, 1, gRGBColors[innercolor]);
		hBrush = CreateSolidBrush(gRGBColors[innercolor]);

        hOldPen = SelectObject(offscreen_DC, hPen);
        hOldBrush = SelectObject(offscreen_DC, hBrush);

        Rectangle(
			offscreen_DC,
            gXCenter + (b->x1+2),
            gYCenter - (b->y1-2),
            gXCenter + (b->x2-2) +1,
            gYCenter - (b->y2+2) +1
		);

		SelectObject(offscreen_DC, hOldPen);
		SelectObject(offscreen_DC, hOldBrush);
		DeleteObject(hPen);
		DeleteObject(hBrush);

		/* Top Bevel */
		hOldPen = SelectObject(offscreen_DC, top_col == WHITE ? hWhitePen : hBlackPen);
		MoveToEx(
			offscreen_DC,
			gXCenter + b->x1,
			gYCenter - b->y1,
			NULL
		);
		LineTo(offscreen_DC, gXCenter + (b->x2-1), gYCenter - (b->y1));
		LineTo(offscreen_DC, gXCenter + (b->x2-2), gYCenter - (b->y1-1));
		LineTo(offscreen_DC, gXCenter + (b->x1+1), gYCenter - (b->y1-1));
		LineTo(offscreen_DC, gXCenter + (b->x1+1), gYCenter - (b->y2+1));
		LineTo(offscreen_DC, gXCenter + (b->x1), gYCenter - (b->y2));
		LineTo(offscreen_DC, gXCenter + (b->x1), gYCenter - (b->y1));
		SelectObject(offscreen_DC, hOldPen);

		/* Bottom Bevel */
		hOldPen = SelectObject(offscreen_DC, bot_col == WHITE ? hWhitePen : hBlackPen);
		MoveToEx(
			offscreen_DC,
			gXCenter + b->x2,
			gYCenter - b->y1,
			NULL
		);
		LineTo(offscreen_DC, gXCenter + (b->x2), gYCenter - (b->y2));
		LineTo(offscreen_DC, gXCenter + (b->x1+1), gYCenter - (b->y2));
		LineTo(offscreen_DC, gXCenter + (b->x1+2), gYCenter - (b->y2+1));
		LineTo(offscreen_DC, gXCenter + (b->x2-1), gYCenter - (b->y2+1));
		LineTo(offscreen_DC, gXCenter + (b->x2-1), gYCenter - (b->y1-1));
		LineTo(offscreen_DC, gXCenter + (b->x2), gYCenter - (b->y1));
		SelectObject(offscreen_DC, hOldPen);


#else
		/* Gray Area */
		GP142_rectangleXY(
			innercolor, 
			b->x1+2, b->y1-2, b->x2-2, b->y2+2, 
			0
		);

		/* Top (white) Bevel */
		GP142_lineXY(top_col, b->x1, b->y1, b->x2-1, b->y1, 1);
		GP142_lineXY(top_col, b->x1, b->y1-1, b->x2-2, b->y1-1, 1);
		GP142_lineXY(top_col, b->x1, b->y1, b->x1, b->y2, 1);
		GP142_lineXY(top_col, b->x1+1, b->y1, b->x1+1, b->y2-1, 1);

		/* Bottom (black) Bevel */
		GP142_lineXY(bot_col, b->x2, b->y1, b->x2, b->y2, 1);
		GP142_lineXY(bot_col, b->x2-1, b->y1-1, b->x2-1, b->y2, 1);
		GP142_lineXY(bot_col, b->x1+2, b->y2+1, b->x2, b->y2+1, 1);
		GP142_lineXY(bot_col, b->x1+1, b->y2, b->x2, b->y2, 1);
#endif
		/* Text */
		GP142_printfXY(b->enabled ? BLACK : MED_GRAY, b->x1+5+shift, (b->y2+b->y1)/2-7-shift, 14, b->label);	

		b->pressed = FALSE;
		b = b->next;
	}

#if defined(WINDOWS) && WIN32_ENHANCED
	DeleteObject(hBlackPen);
	DeleteObject(hWhitePen);
#endif

}


/*
 * Draw the page and all its contents.
 */
void draw_card (instvars_t *ins)
{

	int font_size = 14;
	side_t s = ins->sides[ins->side];

	GP142_printfXY(BLACK, (s.x1 + s.x2)/2 - 50, s.y1 + 5, 14, ins->sidetext);

	if (ins->side == ins->side_to)
	{
		/* No side-flipping animation taking place */
		GP142_rectangleXY (SHADOW_COLOR, s.x1+SHADOW_Z, s.y1-SHADOW_Z, s.x2+SHADOW_Z, s.y2-SHADOW_Z, FILL);
		GP142_rectangleXY (s.color, s.x1, s.y1, s.x2, s.y2, FILL);
		GP142_rectangleXY (BLACK, s.x1, s.y1, s.x2, s.y2, 1);

		/* Draw crease if inside of card */
		if (ins->side == SIDE_INSIDE)
		{
			int cx = (s.x1+s.x2)/2;
			GP142_lineXY (MED_GRAY, cx, s.y1-1, cx, s.y2+1, 1); 
		}

		draw_objects (ins);
	}
	else
	{
		side_t sf, st;  /* Side From, Side To */

		double step;
		int slide;
		double angle;
		int angx, angy;
		int flipcolor, creasecolor, fixedcolor;
		int card_side;

		/* The points of the flipping card */
		GP142_point otr_top, otr_btm, inr_top, inr_btm;

		if (ins->last_gp_event == GP142_PERIODIC) ins->sacnt += FLIP_SPEED;
		if (ins->sacnt > 1.0) ins->sacnt = 1.0;

		if (ins->side == SIDE_FRONT && ins->side_to == SIDE_INSIDE)
		{
			sf = ins->sides[ins->side];
			st = ins->sides[ins->side_to];
			step = ins->sacnt;
			creasecolor = step < 0.5 ? BLACK : MED_GRAY;
			fixedcolor = st.color;
			slide = sf.x1 + (int) ((double) CARD_WIDTH / 2.0 * step);
			card_side = 1;   /* flat card on right */
		}
		else if (ins->side == SIDE_INSIDE && ins->side_to == SIDE_FRONT)
		{
			st = ins->sides[ins->side];
			sf = ins->sides[ins->side_to];
			step = 1.0 - ins->sacnt;
			fixedcolor = st.color;
			creasecolor = step < 0.5 ? BLACK : MED_GRAY;
			slide = sf.x1 + (int) ((double) CARD_WIDTH / 2.0 * step);
			card_side = 1;  /* flat card on right */
		}
		else if (ins->side == SIDE_INSIDE && ins->side_to == SIDE_BACK)
		{
			sf = ins->sides[ins->side];
			st = ins->sides[ins->side_to];
			step = ins->sacnt;
			creasecolor = step < 0.5 ? MED_GRAY : BLACK;
			fixedcolor = sf.color;
			slide = sf.x1 + CARD_WIDTH + (int) ((double) CARD_WIDTH / 2.0 * step);
			card_side = -1;  /* flat card on left */
		}
		else if (ins->side == SIDE_BACK && ins->side_to == SIDE_INSIDE)
		{
			st = ins->sides[ins->side];
			sf = ins->sides[ins->side_to];
			step = 1.0 - ins->sacnt;
			creasecolor = step < 0.5 ? MED_GRAY : BLACK;
			fixedcolor = sf.color;
			slide = sf.x1 + CARD_WIDTH + (int) ((double) CARD_WIDTH / 2.0 * step);
			card_side = -1;  /* flat card on left */
		}

		angle = PI * step;
		angx = (int) ( (double) CARD_WIDTH * cos(angle));
		angy = (int) ( (double) CARD_ZRAISE * sin(angle));
		flipcolor = step < 0.5 ? sf.color : st.color;

		otr_top.x = slide+angx;
		otr_top.y = sf.y1+angy;
		otr_btm.x = slide+angx;
		otr_btm.y = sf.y2-angy;
		inr_top.x = slide;
		inr_top.y = sf.y1;
		inr_btm.x = slide;
		inr_btm.y = sf.y2;

		/* Draw all shadows first */
		GP142_rectangleXY (SHADOW_COLOR, slide+SHADOW_Z, sf.y1-SHADOW_Z, slide+card_side*CARD_WIDTH+SHADOW_Z, sf.y2-SHADOW_Z, FILL);
		GP142_triangleXY (SHADOW_COLOR, inr_top.x+SHADOW_Z, inr_top.y-SHADOW_Z, otr_top.x+SHADOW_Z, otr_top.y-SHADOW_Z, otr_btm.x+SHADOW_Z, otr_btm.y-SHADOW_Z, FILL);
		GP142_triangleXY (SHADOW_COLOR, inr_top.x+SHADOW_Z, inr_top.y-SHADOW_Z, otr_btm.x+SHADOW_Z, otr_btm.y-SHADOW_Z, inr_btm.x+SHADOW_Z, inr_btm.y-SHADOW_Z, FILL);

		/* Fixed card body then outline */
		GP142_rectangleXY (fixedcolor, slide, sf.y1, slide+card_side*CARD_WIDTH, sf.y2, FILL);
		GP142_rectangleXY (BLACK, slide, sf.y1, slide+card_side*CARD_WIDTH, sf.y2, 1);

		/* Flipping card body  */
		GP142_triangleP (flipcolor, inr_top, otr_top, otr_btm, FILL);
		GP142_triangleP (flipcolor, inr_top, otr_btm, inr_btm, FILL);

		/* Flipping card outline */			
		GP142_lineP (BLACK, inr_top, otr_top, 1);
		GP142_lineP (BLACK, otr_top, otr_btm, 1);
		GP142_lineP (BLACK, otr_btm, inr_btm, 1);

		/* Crease (or edge, depending on view */
		GP142_lineXY (creasecolor, slide, sf.y1, slide, sf.y2, 1);

		/* If we're 100% done with the animation, stop the animation */
		if (ins->sacnt == 1.0)
		{
			ins->side = ins->side_to;
			ins->sacnt = 0.0;
			set_status(ins, "");
		}

		if (ins->side_to != ins->side_show && ins->sacnt >= 0.5)
		{
			if (ins->side_to==SIDE_FRONT) assign_string(&ins->sidetext, "Front of Card");
			if (ins->side_to==SIDE_INSIDE) assign_string(&ins->sidetext, "Inside of Card");
			if (ins->side_to==SIDE_BACK) assign_string(&ins->sidetext, "Back of Card");
			ins->side_show = ins->side_to;
		}
	}	

	
	return;
} 

void draw_sel_border (instvars_t *ins)
{
	int i;
	int x1, y1, x2, y2;
	
	if (! ins->sel_object || ins->sel_object->side != ins->side)
		return;

	if (!(ins->sel_object->cx1 || 
	      ins->sel_object->cx2 || 
	      ins->sel_object->cy1 || 
		  ins->sel_object->cy2)
	   )
		return;

	x1 = ins->sel_object->cx1 - SEL_PAD;
	x2 = ins->sel_object->cx2 + SEL_PAD;
	y1 = ins->sel_object->cy1 + SEL_PAD;
	y2 = ins->sel_object->cy2 - SEL_PAD;

	/* This function would be a lot easier if we could use the Win32
	   API and use brushes.  Instead, we draw all the little ants...
	*/
	ins->sel_cycle = (ins->sel_cycle + 1) % 10;
	GP142_rectangleXY(BLACK, x1, y1, x2, y2, 1);

	/* top and bottom antz */
	for (i=ins->sel_cycle; i < x2-x1; i+=10)
	{
		GP142_lineXY(WHITE, x1+i, y1, (x1+i+5 > x2) ? x2 : (x1+i+5), y1, 1);
		GP142_lineXY(WHITE, x2-i, y2, (x2-i-5 < x1) ? x1 : (x2-i-5), y2, 1);
	}
	/* left and right antz */
	for (i=ins->sel_cycle; i < y1-y2; i+=10)
	{
		GP142_lineXY(WHITE, x1, y2+i, x1, (y2+i+5 > y1) ? y1 : (y2+i+5), 1);
		GP142_lineXY(WHITE, x2, y1-i, x2, (y1-i-5 < y2) ? y2 : (y1-i-5), 1);
	}
	/* stupid antz that don't draw normally */
	if (ins->sel_cycle > 4)
	{
		int a = ins->sel_cycle - 4;
		GP142_lineXY(WHITE, x1, y1, x1+a, y1, 1);
		GP142_lineXY(WHITE, x2, y1, x2, y1-1, 1);
		GP142_lineXY(WHITE, x2, y2, x2-a, y2, 1);
		GP142_lineXY(WHITE, x1, y2, x1, y2+a, 1);
	}

}

/*
 * Draw a heart on the card. A heart will be made by two
 * circles. a triangle, which starts on the left circle's
 * PI+PI/4 and the right circles -PI/4 and goes a ways underneith.
 * A rectangle is then used to cover the gap in the middle.
 */
#define HEART_WIDTH 6
void draw_heart (instvars_t *ins, obj_t* obj)
{
	int hmax = obj->size * HEART_WIDTH;
	int h;

	obj->cycle = (obj->cycle + 1) % 12;
	h = hmax - (int)(0.6 * (double)HEART_WIDTH * sin((double) obj->cycle * PI / 12.0));

	GP142_circleXY (obj->color, obj->x - h, obj->y, h);
	GP142_circleXY (obj->color, obj->x + h, obj->y, h);
	GP142_triangleXY (
		obj->color, 
		obj->x-h-(int)((double)h*COS_HALF_PI), obj->y-(int)((double)h*COS_HALF_PI)-1,
		obj->x+h+(int)((double)h*COS_HALF_PI), obj->y-(int)((double)h*COS_HALF_PI)-1,
		obj->x, obj->y-h*3,				
		FILL
	);
	GP142_rectangleXY (
		obj->color, 
		obj->x-h, obj->y,
		obj->x+h, obj->y-h,
		FILL		
	);

	if (obj->recalc_coords)
		set_object_coords(
			obj, 
			obj->x-hmax*2, obj->y+hmax,
			obj->x+hmax*2, obj->y-hmax*3
		);
}

void draw_star (instvars_t *ins, obj_t *obj)
{
	int i;
	int x[10], y[10];
	int radius = obj->size * 12;
	double angle;

	obj->cycle = (obj->cycle + 1) % 10;
	angle = PI / 5.0 * (double) obj->cycle / 5.0;
	for (i=0; i<10; i++)
	{
		x[i] = obj->x + (int) (((i%2) ? 1.0 : 0.4)  * radius * cos(angle));
		y[i] = obj->y + (int) (((i%2) ? 1.0 : 0.4)  * radius * sin(angle));
		angle += PI / 5.0;
	}	

	for (i=0; i<=8; i+=2)
	{
		GP142_triangleXY (obj->color, x[i], y[i], x[i+1], y[i+1], x[(i+2)%10], y[(i+2)%10], FILL);
	}
	GP142_triangleXY (obj->color, x[2], y[2], x[4], y[4], x[6], y[6], FILL);
	GP142_triangleXY (obj->color, x[2], y[2], x[8], y[8], x[6], y[6], FILL);
	GP142_triangleXY (obj->color, x[2], y[2], x[0], y[0], x[8], y[8], FILL);

	if (obj->recalc_coords)
		set_object_coords(
			obj, 
			obj->x-radius, obj->y+radius,
			obj->x+radius, obj->y-radius
		);
}

void draw_balloon (instvars_t *ins, obj_t *obj)
{
	double d;
	int lx=0, nx, ly, ny;
	int radx = obj->size * 10;
	int rady = obj->size * 15;
	double theta, tchange;

	obj->cycle = (obj->cycle + 1) % 12;
	theta = PI / 2.0;
	tchange = (obj->cycle - 5) / 9.0;

	for (d=0.0; d<12.0; d++)
	{
		nx = (int) (obj->x + obj->size * 4 * cos(theta));
		ny = obj->y - rady - (int)((rady * 2) * d/12.0);
		if (lx)
			GP142_lineXY (
				BLACK, 
				lx, ly,
				nx, ny, 
				1
			);
		lx = nx;
		ly = ny;
		theta += tchange;
	}

	GP142_ovalXY (
		obj->color, 
		obj->x - radx, obj->y + rady, 
		obj->x + radx,	obj->y - rady + obj->size*2,
		FILL
	);
	GP142_triangleXY (
		obj->color,
		obj->x - radx/2, obj->y - rady,
		obj->x + radx/2, obj->y - rady,
		obj->x, obj->y,
		FILL
	);

	if (obj->recalc_coords)
		set_object_coords(
			obj, 
			obj->x-radx, obj->y+rady,
			obj->x+radx, obj->y-rady*3
		);

}

void draw_dub (instvars_t *ins, obj_t *obj)
{
	double wu, wmax = obj->size * 2;
	double hu = obj->size * 3;

	int i;

	obj->cycle = (obj->cycle + 1) % 46;
	wu = wmax * sin(2.0 * (obj->cycle / 48.0) * PI);

	for (i=-1; i<=1; i+=2)
	{
		/* Working from the inside triangles out, mirror on both sides */
		GP142_triangleXY
		(
			obj->color,
			obj->x, (int) (obj->y + hu * 3),
			obj->x, (int) (obj->y - hu * 2),
			(int) (obj->x + i*wu*3), (int) (obj->y - hu * 2),
			FILL
		);

		GP142_triangleXY
		(
			obj->color,
			obj->x, (int) (obj->y - hu * 2),
			(int) (obj->x - i * wu * 1), (int) (obj->y - hu * 4),
			(int) (obj->x - i * wu * 3), (int) (obj->y - hu * 2),
			FILL
		);

		GP142_triangleXY
		(
			obj->color,
			(int) (obj->x - i * wu * 5), (int) (obj->y - hu * 4),
			(int) (obj->x - i * wu * 1), (int) (obj->y - hu * 4),
			(int) (obj->x - i * wu * 3), (int) (obj->y - hu * 2),
			FILL
		);

		GP142_triangleXY
		(
			obj->color,
			(int) (obj->x - i * wu * 5), (int) (obj->y + hu * 2),
			(int) (obj->x - i * wu * 5), (int) (obj->y - hu * 4),
			(int) (obj->x - i * wu * 3), (int) (obj->y - hu * 2),
			FILL
		);

		GP142_triangleXY
		(
			obj->color,
			(int) (obj->x - i * wu * 8), (int) (obj->y + hu * 2),
			(int) (obj->x - i * wu * 5), (int) (obj->y + hu * 2),
			(int) (obj->x - i * wu * 5), (int) (obj->y - hu * 4),
			FILL
		);

		GP142_rectangleXY
		(
			obj->color,
			(int) (obj->x - i * wu * 10), (int) (obj->y + hu * 4),
			(int) (obj->x - i * wu * 3), (int) (obj->y + hu * 2),
			FILL
		);
	}
	if (obj->recalc_coords)
		set_object_coords(
			obj, 
			(int) (obj->x - wmax * 10), (int) (obj->y + hu * 4),
			(int) (obj->x + wmax * 10), (int) (obj->y - hu * 4)
		);
}

void handle_event (instvars_t *ins, int event, int subevent, int mouse_x, int mouse_y)
{
	switch(event)
	{
		case 0:
			break;

		case EVENT_EMAIL:
#if defined(WINDOWS) && WIN32_ENHANCED && EMAIL_SHELL_EXECUTE
			ShellExecute(hDrawWnd, "open", "mailto:bradfitz@bradfitz.com", NULL, NULL, 1);
#endif
			break;

		case EVENT_CHANGE_COLOR:
			if (ins->sel_object)
				ins->sel_object->color = (ins->sel_object->color + 1) % MAX_COLORS;
			break;

		case EVENT_CLONE:
			if (ins->sel_object)
			{
				obj_t *obnew;
				int cx, cy;
				double angle;

				cx = (ins->sides[ins->side].x1 + ins->sides[ins->side].x1) / 2;
				cy = (ins->sides[ins->side].y1 + ins->sides[ins->side].y1) / 2;

				obnew = clone_object(ins->sel_object);

				obnew->to_x = ins->sides[ins->side].x1 + rand() % (ins->sides[ins->side].x2-ins->sides[ins->side].x1);
				obnew->to_y = ins->sides[ins->side].y1 - rand() % (ins->sides[ins->side].y1-ins->sides[ins->side].y2);
				
				angle = atan2(obnew->to_y-CARD_MY, obnew->to_x-CARD_MX);
				obnew->x = obnew->from_x = (int) (GP142_XMAX * cos(angle));
				obnew->y = obnew->from_y = (int) (GP142_XMAX * sin(angle));
				obnew->acnt = 0.0;

				obnew->cycle = rand() % 100;  /* doesn't matter what, just something random */
				obnew->recalc_coords = 1;

				add_object(ins, obnew);
				ins->sel_object = obnew;
			}
			break;

		case EVENT_CHANGE_OBJECT:
			if (ins->sel_object)
			{
				if (ins->sel_object->type == OBJ_TEXT)
				{
					char message[MAX_STRING_SIZE];
					GP142_gets("Enter new text.", message);
					printf("string = %s", message);
					if (strlen(message))
						assign_string(&ins->sel_object->text, message);
					ins->sel_object->recalc_coords = 1;
				
				}
				else if (ins->sel_object->type == OBJ_SHAPE)
				{
					ins->sel_object->shape = (ins->sel_object->shape + 1) % MAX_SHAPES;
					ins->sel_object->recalc_coords = 1;
				}

			}
			break;

		case EVENT_DELETE:
			if (ins->sel_object && ins->sel_object->type != OBJ_EXPLOSION)
			{
				ins->sel_object->type = OBJ_EXPLOSION;
				ins->sel_object->cycle = 0;
				ins->sel_object->cx1 = ins->sel_object->cx2 = 0;
				ins->sel_object->cy1 = ins->sel_object->cy2 = 0;

			}
			ins->sel_object = NULL;  /* De-select */
			ins->status = STATUS_NONE;
			break;

		case EVENT_KILL_ALL:
			if (ins->side != ins->side_to) break;

			/* Using sel_object as a temp variable, but it doesn't matter
			 * cause it gets destroyed later anyway.
			 */
			ins->sel_object = ins->objects->first;
			while (ins->sel_object)
			{
				if (ins->sel_object->side == ins->side && 
					ins->sel_object->type != OBJ_EXPLOSION)
				{
					ins->sel_object->type = OBJ_EXPLOSION;
					ins->sel_object->cycle = 0;
					ins->sel_object->cx1 = ins->sel_object->cx2 = 0;
					ins->sel_object->cy1 = ins->sel_object->cy2 = 0;
				}
				ins->sel_object = ins->sel_object->next;
			}
			ins->status = STATUS_NONE;
			break;

		case EVENT_ADD_OBJECT:
			ins->status = ins->status != STATUS_ADD_OBJECT ? STATUS_ADD_OBJECT : 0;
			set_status(ins, "ADD SHAPE: Click on card to add a shape");
			break;

		case EVENT_MOVE:
			ins->status = ins->status != STATUS_MOVE ? STATUS_MOVE : 0;
			set_status(ins, "MOVE OBJECT: Click on card to move selected object to that position.");
			break;

		case EVENT_ADD_TEXT:
			ins->status = ins->status != STATUS_ADD_TEXT ? STATUS_ADD_TEXT : 0;
			set_status(ins, "ADD TEXT: Click on card to add a message");
			break;

		case EVENT_SELECT:
			ins->sel_object = NULL;
			ins->status = ins->status != STATUS_SELECT ? STATUS_SELECT : 0;
			set_status(ins, "SELECT OBJECT: Click near middle of shape or lower-left of text object");
			break;

		case EVENT_CHANGE_SIZE:
			if (ins->sel_object)
			{
				if (subevent == 1 && ins->sel_object->size < SIZE_LARGE)
					ins->sel_object->size++;
				if (subevent == -1 && ins->sel_object->size > SIZE_SMALL)
					ins->sel_object->size--;

				ins->sel_object->recalc_coords = 1;
			}
			break;

		case EVENT_CHANGE_BG:
			ins->status = STATUS_NONE;
			/* Change the color of the card we're flipping to (or on, if stopped) */
			ins->sides[ins->side_to].color = (ins->sides[ins->side_to].color + 1) % MAX_COLORS;
			break;

		case EVENT_COLOR:
			/* If an object is selected, change that color.  Else, change background color */
			switch (subevent)
			{
				case 0: set_status(ins, "Black"); break;
				case 1: set_status(ins, "White"); break;
				case 2: set_status(ins, "Brick Red"); break;
				case 3: set_status(ins, "Kermit Green"); break;
				case 4: set_status(ins, "Deep Blue"); break;
				case 5: set_status(ins, "Schoolbus Yellow"); break;
				case 6: set_status(ins, "Lavender"); break;
				case 7: set_status(ins, "Turquriosoisoie"); break;
				case 8: set_status(ins, "Purple"); break;
				case 9: set_status(ins, "Navy Blue"); break;
				case 10: set_status(ins, "Grape"); break;
				case 11: set_status(ins, "Light Turquoioises"); break;
				case 12: set_status(ins, "Light Sea Green"); break;
				case 13: set_status(ins, "Pumpkin Pie Orange"); break;
				case 14: set_status(ins, "Shit Brown"); break;
				case 15: set_status(ins, "Bubblegum Pink"); break;
				case 16: set_status(ins, "Popcorn Yellow"); break;
				case 17: set_status(ins, "Coke Slurpee Brown"); break;
				case 18: set_status(ins, "Puke Orange"); break;
				case 19: set_status(ins, "Money Green"); break;
				case 20: set_status(ins, "Mint Green"); break;
				case 21: set_status(ins, "Rotten Avacado Green"); break;
				case 22: set_status(ins, "Dark Gray"); break;
				case 23: set_status(ins, "Light Gray"); break;
			}
			if (ins->sel_object)
				ins->sel_object->color = subevent;
			else
				ins->sides[ins->side_to].color = subevent;
			break;

		case EVENT_FLIP_SIDE:
			ins->status = STATUS_NONE;
			ins->sel_object = NULL;  /* De-select objects on side flip */

			if (ins->side == ins->side_to)
			{
				/* If not already in motion.... */
				if (subevent == -1 && ins->side == SIDE_FRONT)
					ins->side_to = SIDE_INSIDE;
				else if (subevent == -1 && ins->side == SIDE_INSIDE)
					ins->side_to = SIDE_BACK;

				if (subevent == 1 && ins->side == SIDE_INSIDE)
					ins->side_to = SIDE_FRONT;
				else if (subevent == 1 && ins->side == SIDE_BACK)
					ins->side_to = SIDE_INSIDE;
			}
			else
			{
				if (ins->side == SIDE_FRONT && ins->side_to == SIDE_INSIDE && subevent == 1)
				{
					ins->side = SIDE_INSIDE;
					ins->side_to = SIDE_FRONT;
					ins->sacnt = 1.0 - ins->sacnt;
				}
				else if (ins->side == SIDE_INSIDE && ins->side_to == SIDE_FRONT && subevent == -1)
				{
					ins->side = SIDE_FRONT;
					ins->side_to = SIDE_INSIDE;
					ins->sacnt = 1.0 - ins->sacnt;
				}
				else if (ins->side == SIDE_INSIDE && ins->side_to == SIDE_BACK && subevent == 1)
				{
					ins->side = SIDE_BACK;
					ins->side_to = SIDE_INSIDE;
					ins->sacnt = 1.0 - ins->sacnt;
				}
				else if (ins->side == SIDE_BACK && ins->side_to == SIDE_INSIDE && subevent == -1)
				{
					ins->side = SIDE_INSIDE;
					ins->side_to = SIDE_BACK;
					ins->sacnt = 1.0 - ins->sacnt;
				}
			}

			break;
	}

	if (event)
	{
		if (event != EVENT_COLOR && ins->status == STATUS_NONE)
			set_status(ins, "");
		return;
	}
}

/*
 * Decide where a mouse click occured, and react appropriately.
 */
void handle_mouse_click(int mouse_x, int mouse_y, instvars_t *ins)
{
	int event = 0;
	int subevent = 0;
	side_t s;

	handle_button(mouse_x, mouse_y, &event, &subevent, ins);
	handle_event (ins, event, subevent, mouse_x, mouse_y);

	/* Ignore click if we're in the middle of flipping the card. */
	if (ins->side != ins->side_to) return;

	s = ins->sides[ins->side];

	/* Add Shape or Text */
	if ((ins->status == STATUS_ADD_OBJECT || ins->status == STATUS_ADD_TEXT) &&
	    mouse_x > s.x1 && mouse_x < s.x2 && mouse_y < s.y1 && mouse_y > s.y2)
	{
		obj_t *newobj = create_object(mouse_x, mouse_y, ins);
		add_object(ins, newobj);
		ins->status = STATUS_NONE;
		ins->sel_object = newobj;
		set_status(ins, "New object selected.");
		return;
	}

	/* Move Selected Object */
	if ((ins->status == STATUS_MOVE && ins->sel_object) &&
	    mouse_x > s.x1 && mouse_x < s.x2 && mouse_y < s.y1 && mouse_y > s.y2)
	{
		ins->sel_object->from_x = ins->sel_object->x;
		ins->sel_object->from_y = ins->sel_object->y;
		ins->sel_object->to_x = mouse_x;
		ins->sel_object->to_y = mouse_y;
		ins->sel_object->acnt = 0.0;
		return;
	}

	/* Select Object (default action) */
	if (mouse_x > s.x1 && mouse_x < s.x2 && mouse_y < s.y1 && mouse_y > s.y2)
	{
		obj_t *newsel = NULL;
		int closest = GP142_XMAX * 2;  /* something ridiculuously far away */
		int thisdist;
		obj_t *o = ins->objects->first;

		while (o)
		{
			if (o->side == ins->side)
			{
				thisdist = (int) sqrt((o->x-mouse_x)*(o->x-mouse_x)+(o->y-mouse_y)*(o->y-mouse_y));
				if (thisdist < closest)
				{
					closest = thisdist;
					newsel = o;
				}
			}
			o = o->next;
		}

		/* If clicked one is farther than SELECT_RADIUS away AND it isn't
		   within the shape's boundaries, deselect it.  Nothing was caught.  */
		if (newsel && closest > SELECT_RADIUS && (
		    mouse_x < newsel->cx1 || mouse_x > newsel->cx2 ||
			mouse_y > newsel->cy1 || mouse_y < newsel->cy2))
		{
			newsel = NULL;
		}

		if (ins->sel_object == newsel)
		{
			ins->sel_object = NULL;
			ins->status = STATUS_NONE;
			set_status(ins, "");
		}
		else
		{
			ins->sel_object = newsel;
			ins->status = STATUS_NONE;  /* done selecting, we found one */
			if (newsel) 
				set_status(ins, "Object Selected.");
			else
				set_status(ins, "");
		}
		return;
	}
	
	return;
}

/* 
 * Figure out which button was clicked, if any, and do the right 
 * thing.
 */
void handle_button(int x, int y, int *event, int *subevent, instvars_t *ins)
{
	button_t *b = ins->buttons->first;

	while (b && ! *event)
	{
		if (b->enabled && x >= b->x1 && x <= b->x2 && y <= b->y1 && y >= b->y2)
		{
			*event = b->event;
			*subevent = b->subevent;
			b->pressed = TRUE;
		}
		b = b->next;
	}
}

obj_t* clone_object(obj_t* source)
{
	obj_t *dest;

	dest = NEW(obj_t);
	dest->next = NULL;
	dest->prev = NULL;

	dest->type = source->type;
	dest->shape = source->shape;
	dest->cycle = source->cycle;
	dest->color = source->color;
	dest->size = source->size;
	dest->recalc_coords = source->recalc_coords;

	dest->cx1 = source->cx1;
	dest->cy1 = source->cy1;
	dest->cx2 = source->cx2;
	dest->cy2 = source->cy2;

	assign_string(&dest->text, source->text);  

	dest->x = source->x;
	dest->y = source->y;
	dest->side = source->side;

	dest->from_x = source->from_x;
	dest->from_y = source->from_y;
	dest->to_x = source->to_x;
	dest->to_y = source->to_y;
	dest->acnt = source->acnt;

	return dest;
}

obj_t* create_object(int x, int y, instvars_t *ins)
{
	double angle = atan2(y-CARD_MY, x-CARD_MX);
	obj_t *obj;

	obj = NEW(obj_t);
	obj->next = NULL;
	obj->prev = ins->objects->last;
	obj->type = ins->status == STATUS_ADD_TEXT ? OBJ_TEXT : OBJ_SHAPE;
	obj->shape = SHAPE_HEART;
	obj->color = RED;
	obj->text = NULL;
	obj->side = ins->side_to;
	obj->to_x = x; 
	obj->from_x = obj->x = (int) (GP142_XMAX * cos(angle));
	obj->to_y = y; 
	obj->from_y = obj->y = (int) (GP142_XMAX * sin(angle));
	obj->acnt = 0.0;	
	obj->size = SIZE_SMALL_MED;
	obj->recalc_coords = 1;
	obj->cycle = 0;

	if (ins->status == STATUS_ADD_TEXT)
	{
		char message[MAX_STRING_SIZE];
		GP142_gets("What message would you like to insert?", message);

		if (strlen(message))
			assign_string(&obj->text, message);
		else
			assign_string(&obj->text, DEFAULT_MESSAGE);
	}

	return obj;
}

void add_object (instvars_t *ins, obj_t *obj)
{
	ins->objects->count++;
	if (ins->objects->count == 1)
	{
		ins->objects->first = obj;
		ins->objects->last = obj;
		obj->prev = NULL;
	}
	else
	{
		ins->objects->last->next = obj;
		obj->prev = ins->objects->last;
		ins->objects->last = obj;
	}

	return;
}

void draw_object (instvars_t *ins, obj_t *obj)
{
	if (obj->x != obj->to_x || obj->y != obj->to_y)
	{
		obj->acnt += MOVE_SPEED;
		if (obj->acnt >= 1.0)
		{
			obj->x = obj->to_x;
			obj->y = obj->to_y;
		}
		else
		{
			double percent = sin(PI/2.0*obj->acnt);
			obj->x = obj->from_x + (int) ((double)(obj->to_x - obj->from_x) * percent);
			obj->y = obj->from_y + (int) ((double)(obj->to_y - obj->from_y) * percent);
		}
		obj->recalc_coords = 1;
	}

	/* An object's type is OBJ_EXPLOSION if it's been deleted.  It's actually deleted
	 * when it's done exploding.  
	 */
	if (obj->type == OBJ_EXPLOSION)
	{
		int inc;
		inc = GP142_XMAX/10;

		if (obj->cycle > inc*5)
		{
			delete_object (ins, obj);
		}
		else
		{
			int i;
			double ang;
			
			obj->cycle += inc;
			for (i=0; i<20; i++)
			{
				ang = 2.0 * PI * (double) i / 20.0;
				if (i%2) obj->cycle += inc/2;
				GP142_lineXY (
					obj->color,
					obj->x + (int) (obj->cycle * cos(ang)), 
					obj->y + (int) (obj->cycle * sin(ang)),
					obj->x + (int) ((obj->cycle+inc) * cos(ang)), 
					obj->y + (int) ((obj->cycle+inc) * sin(ang)),
					obj->cycle < inc*3 ? 2 : 1
				);
				if (i%2) obj->cycle -= inc/2;
			}
		}
		return;
	}

	if (obj->type == OBJ_SHAPE)
	{
		switch (obj->shape)
		{
			case SHAPE_HEART:
				draw_heart (ins, obj);
				break;

			case SHAPE_STAR:
				draw_star (ins, obj);
				break;

			case SHAPE_BALLOON:
				draw_balloon (ins, obj);
				break;

			case SHAPE_DUB:
				draw_dub (ins, obj);
				break;
		}
	}

	if (obj->type == OBJ_TEXT)
	{
		if (ins->sel_object == obj && ins->sel_object->recalc_coords)
		{
			int width, height;
			text_size (&width, &height, get_fontsize(obj->size), obj->text);
			set_object_coords (obj, obj->x-width/2, obj->y+height/2, obj->x+width/2, obj->y-height/2);
		}
		GP142_printfXY(
			obj->color, 
			obj->x-(obj->cx2-obj->cx1)/2, 
			obj->y-(obj->cy1-obj->cy2)/2, 
			get_fontsize(obj->size), 
			obj->text
		);
	}

}

void draw_objects (instvars_t *ins)
{
	obj_t *obj = ins->objects->first;

	while (obj)
	{
		if (obj->side == ins->side)
			draw_object (ins, obj);

		obj = obj->next;
	}
}